package java.util.stream;

import java.util.concurrent.ForkJoinPool;
import java.util.Spliterator;
import java.util.concurrent.CountedCompleter;

abstract class AbstractTask<P_IN, P_OUT, R, K extends AbstractTask<P_IN, P_OUT, R, K>> extends CountedCompleter<R>
{
    static final int LEAF_TARGET;
    protected final PipelineHelper<P_OUT> helper;
    protected Spliterator<P_IN> spliterator;
    protected long targetSize;
    protected K leftChild;
    protected K rightChild;
    private R localResult;
    
    protected AbstractTask(PipelineHelper<P_OUT> helper, Spliterator<P_IN> spliterator) {
        super(null);
        this.helper = helper;
        this.spliterator = spliterator;
        this.targetSize = 0L;
    }
    
    protected AbstractTask(K completer, Spliterator<P_IN> spliterator) {
        super(completer);
        this.spliterator = spliterator;
        this.helper = completer.helper;
        this.targetSize = completer.targetSize;
    }
    
    protected abstract K makeChild(Spliterator<P_IN> p0);
    
    protected abstract R doLeaf();
    
    public static long suggestTargetSize(long n) {
        long n2 = n / (long)AbstractTask.LEAF_TARGET;
        return (n2 > 0L) ? n2 : 1L;
    }
    
    protected final long getTargetSize(long n) {
        long targetSize;
        return ((targetSize = this.targetSize) != 0L) ? targetSize : (this.targetSize = suggestTargetSize(n));
    }
    
    @Override
    public R getRawResult() {
        return this.localResult;
    }
    
    @Override
    protected void setRawResult(R r) {
        if (r != null) {
            throw new IllegalStateException();
        }
        return;
    }
    
    protected R getLocalResult() {
        return this.localResult;
    }
    
    protected void setLocalResult(R localResult) {
        this.localResult = localResult;
    }
    
    protected boolean isLeaf() {
        return this.leftChild == null;
    }
    
    protected boolean isRoot() {
        return this.getParent() == null;
    }
    
    protected K getParent() {
        return (AbstractTask)this.getCompleter();
    }
    
    @Override
    public void compute() {
        Spliterator<P_IN> spliterator = this.spliterator;
        long n = spliterator.estimateSize();
        long targetSize = this.getTargetSize(n);
        int n2 = 0;
        Spliterator<P_IN> trySplit;
        while (n > targetSize && (trySplit = spliterator.trySplit()) != null) {
            AbstractTask<Object, Object, Object, Object> abstractTask = this.leftChild = this.makeChild(trySplit);
            AbstractTask<Object, Object, Object, Object> abstractTask2 = this.rightChild = this.makeChild(spliterator);
            this.setPendingCount(1);
            if (n2 != 0) {
                n2 = 0;
                spliterator = trySplit;
                this = abstractTask;
            }
            else {
                n2 = 1;
                this = abstractTask2;
                abstractTask2 = abstractTask;
            }
            abstractTask2.fork();
            n = spliterator.estimateSize();
        }
        this.setLocalResult(this.doLeaf());
        this.tryComplete();
    }
    
    @Override
    public void onCompletion(CountedCompleter<?> countedCompleter) {
        this.spliterator = null;
        AbstractTask<P_IN, P_OUT, R, K> abstractTask = null;
        this.rightChild = abstractTask;
        this.leftChild = abstractTask;
    }
    
    protected boolean isLeftmostNode() {
        while (this != null) {
            AbstractTask<P_IN, P_OUT, R, AbstractTask<P_IN, P_OUT, R, K>> parent = this.getParent();
            if (parent != null && parent.leftChild != this) {
                return false;
            }
            this = parent;
        }
        return true;
    }
    
    static {
        LEAF_TARGET = ForkJoinPool.getCommonPoolParallelism() << 2;
    }
}
